# sftprepo.py - sftp URL scheme for accessing repositories
#
# Copyright (c) 2009-2012 G.P. Halkes <sftprepo@ghalkes.nl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# sftprepo version 0.1.4

# Current limitations:
#
# 1) If you want to create a new mercurial repository on the remote end, you
# first have to create all parent directories. This is because the mount uses
# the parent directory of the repository as the remote directory to mount.
#
# 2) No options can be passed to the mount command from the command line.

'''sftp URL scheme for accessing repositories'''

from mercurial import hg, localrepo, util, repo, error
import sys, tempfile, os, re, atexit, weakref

# Maintain a list of weak references to the sftprepository objects. If the
# object is still alive at exit, it is close'd, such that the repository is
# unmounted.
__exitHandlerRegistered__ = False
__openRepos__ = []
def __closeRepos__():
	for repo in __openRepos__:
		repoObject = repo()
		if repoObject:
			repoObject.close()

class sftprepository(localrepo.localrepository):
	def __init__(self, parentui, path=None, create=0):
		global __exitHandlerRegistered__, __openRepos__
		if not __exitHandlerRegistered__:
			atexit.register(__closeRepos__)
			__exitHandlerRegistered__ = True
		self.mountdir = tempfile.mkdtemp('', 'sftprepo')
		# Because the mount directory gets the rights of the mounted
		# directory, make a subdirectory such that no-one else can touch the
		# mounted dir
		self.fullmountdir = self.mountdir + "/mnt"
		os.mkdir(self.fullmountdir)
		__openRepos__.append(weakref.ref(self))

		# Take the URL apart
		parts = re.match('sftp://([^/]+)/((.+/)?)([^/]+)/?', path)

		if parts is None:
			os.rmdir(self.fullmountdir)
			os.rmdir(self.mountdir)
			self.mountdir = None
			raise error.RepoError("Invalid sftp URL: %s" % path)

		remotedir = parts.group(1) + ':' + parts.group(2)

		# Mount the remote directory so we can access it as if it is local
		# Don't know if the rename workaround is needed, but it is for git so
		# just assume it is required.
		if self.dosystem(["sshfs", "-o", "workaround=rename", remotedir, self.fullmountdir]) != 0:
			os.rmdir(self.fullmountdir)
			os.rmdir(self.mountdir)
			self.mountdir = None
			raise error.RepoError("Error mounting remote repository")

		# Setup "local" repository in the mounted directory
		localrepo.localrepository.__init__(self, parentui,
			self.fullmountdir + "/" + parts.group(4), create)

	def dosystem(self, args):
		try:
			pid = os.fork()
			if pid == 0:
				os.execvp(args[0], args)
			else:
				(_pid, retval) = os.waitpid(pid, 0)
			return retval
		except:
			return -1

	def _cleanup(self):
		# Unmount and remove the mount point
		if not self.mountdir is None:
			self.dosystem(["fusermount", "-u", self.fullmountdir])
			os.rmdir(self.fullmountdir)
			os.rmdir(self.mountdir)
			self.mountdir = None

	def __del__(self):
		self._cleanup()

	def close(self):
		super(sftprepository, self).close()
		self._cleanup()

	def local(self):
		# Although we are a subclass of localrepo, we are not actually a
		# local repository
		return False

def instance(ui, path, create):
	return sftprepository(ui, path, create)

# Add sftp to the list of supported schemes
hg.schemes['sftp'] = sys.modules[__name__]
